#ifndef _LIGO_FILTERIO_H
#define _LIGO_FILTERIO_H
/*----------------------------------------------------------------------*/
/*                                                         		*/
/* Module Name: FilterFile						*/
/*                                                         		*/
/* Module Description: Filter IO to read and write online files		*/
/*                                                         		*/
/* Revision History:					   		*/
/* Rel   Date     Programmer  	Comments				*/
/* 1.0	 4Aug02   D. Sigg    	First release		   		*/
/*                                                         		*/
/* Documentation References:						*/
/*	Man Pages: FilterFile.html					*/
/*	References: none						*/
/*                                                         		*/
/* Author Information:							*/
/* Name          Telephone       Fax             e-mail 		*/
/* Daniel Sigg   (509) 372-8132  (509) 372-8137  sigg_d@ligo.mit.edu	*/
/*                                                         		*/
/*                                                         		*/
/*                      -------------------                             */
/*                                                         		*/
/*                             LIGO					*/
/*                                                         		*/
/*        THE LASER INTERFEROMETER GRAVITATIONAL WAVE OBSERVATORY.	*/
/*                                                         		*/
/*                     (C) The LIGO Project, 1999.			*/
/*                                                         		*/
/*                                                         		*/
/* Caltech				MIT		   		*/
/* LIGO Project MS 51-33		LIGO Project NW-17 161		*/
/* Pasadena CA 91125			Cambridge MA 01239 		*/
/*                                                         		*/
/* LIGO Hanford Observatory		LIGO Livingston Observatory	*/
/* P.O. Box 1970 S9-02			19100 LIGO Lane Rd.		*/
/* Richland WA 99352			Livingston, LA 70754		*/
/*                                                         		*/
/*----------------------------------------------------------------------*/

#include "FilterDesign.hh"
#include "filterwiz/FilterSection.hh"
#include "filterwiz/FilterModule.hh"
#include <string>
#include <vector>


namespace filterwiz {


   /// Magic string for an online filter file
   const char* const kMagic = "# FILTERS FOR ONLINE SYSTEM";


/** Filter File Class. Reads and writes the filter files of the online
    system.
   
    @memo Filter File IO
    @author Written August 2002 by Daniel Sigg
    @version 1.0
 ************************************************************************/
   class FilterFile {
   public:
      /// Default constructor
      FilterFile();
   
      /// Clear all modules from list
      void clear();
      /// Add a module
      void add (const char* name, double fsample = 1.0);
      /// Remove a module
      void remove (const char* name);
      /// Find module
      FilterModule* find (const char* name);
      /// Find module
      const FilterModule* find (const char* name) const;
      /// Filter module list
      const FilterModuleList& modules() const {
         return fModules; }
      /// Filter module list
      FilterModuleList& modules() {
         return fModules; }
      /// Get filename
      const char* getFilename() const {
         return fFilename.c_str(); }
   
      /// check if filters are valid
      bool valid (std::string& errmsg) const;
      /// update filters
      bool update();
   
      /// Open and read a online file
      bool read (const char* filename);
      /// Read a module from a char array
      int read (const char* p, int maxlen);
      /// Write a online file
      bool write (const char* filename);
      /// Write a module to a char array
      int write (char* p, int maxlen);
   
   
   protected:
      /// Filter file
      std::string		fFilename;
      /// Filter module list
      FilterModuleList		fModules;
   
      /// Returns a valid filter section name
      std::string validSectionName (const char* name);
   };

}

#endif // _LIGO_FILTERIO_H
