#ifndef _LIGO_FILTERDESIGN_H
#define _LIGO_FILTERDESIGN_H
/*----------------------------------------------------------------------*/
/*                                                         		*/
/* Module Name: FilterDesign						*/
/*                                                         		*/
/* Module Description: design filters					*/
/*                                                         		*/
/* Revision History:					   		*/
/* Rel   Date     Programmer  	Comments				*/
/* 0.1	 17Jul02  D. Sigg    	First release		   		*/
/*                                                         		*/
/* Documentation References:						*/
/*	Man Pages: FilterDesign.html					*/
/*	References: none						*/
/*                                                         		*/
/* Author Information:							*/
/* Name          Telephone       Fax             e-mail 		*/
/* Daniel Sigg   (509) 372-8132  (509) 372-8137  sigg_d@ligo.mit.edu	*/
/*                                                         		*/
/*                                                         		*/
/*                      -------------------                             */
/*                                                         		*/
/*                             LIGO					*/
/*                                                         		*/
/*        THE LASER INTERFEROMETER GRAVITATIONAL WAVE OBSERVATORY.	*/
/*                                                         		*/
/*                     (C) The LIGO Project, 1999.			*/
/*                                                         		*/
/*                                                         		*/
/* Caltech				MIT		   		*/
/* LIGO Project MS 51-33		LIGO Project NW-17 161		*/
/* Pasadena CA 91125			Cambridge MA 01239 		*/
/*                                                         		*/
/* LIGO Hanford Observatory		LIGO Livingston Observatory	*/
/* P.O. Box 1970 S9-02			19100 LIGO Lane Rd.		*/
/* Richland WA 99352			Livingston, LA 70754		*/
/*                                                         		*/
/*----------------------------------------------------------------------*/

#include "Complex.hh"
#include "Pipe.hh"
#include "TSeries.hh"
#include "FSeries.hh"
#include "FilterParse.hh"
#include <string>

   class Chirp;
   class SweptSine;
namespace ligogui {
   class TLGMultiPad;
}

/** Design a filter.
    The filter design class allows to design complex filters. It is
    not a filter class by itself but rather a "factory" class that
    produces filters depending on a user specifiaction.
    It supports:
    \begin{verbatim}
    1. IIR: elliptic, chebyshev, butterworth, zero/pole/gain, notches,
       resonant gain, pedndulum and biquad coefficients.
    2. FIR: window and McClellan-Parks algorithms.
    3. Others: line tracker, decimation and resampling.
    \end{verbatim}
    Any number of filters can be added in series. Filters can be
    designed by calling their corresponding creation methods (such
    as ellip), by parsing a string formula or by reading the 
    coefficients of a configuration file used by the front-end code.

    The filter design class also supports a configuration line 
    parser which follows the matlab syntax. This way a filter can 
    be written by a product of designer functions. For example,
    "zpk([100;100],[1;1],1) /. setgain(0,1)" represents a filter
    consisting of two poles at 100Hz, two zeros at 1Hz and a 
    dc gain of 1. For a list of supported designer functions
    see below.

    For filters that support it such as IIR, FIR and combinations
    of them the filter design class can also calculate the 
    transfer function and generate a Bode plot. All filters can
    tested for step, impulse and ramp response.

    By default the filter design class will return a unity gain 
    filter (identity operator), if no filter has been added.

    \begin{verbatim}
    Example 1: Make an elliptic high pass filter
    FilterDesign fdesign (16384); // specify sampling frequency
    fdesign.ellip (kLowPass,8,1,60,100); // make elliptic low pass
           // 8th order, 1dB ripple, 60dB attenuation, 100Hz corner
    Pipe* pipe = fdesign.release(); // get filter
    \end{verbatim}

    \begin{verbatim}
    Example 2: Make a whitening filter with a resonant gain, generate 
               a bode plot and investigate its step response
    FilterDesign fdesign ("zpk([1;1],[100;100],1)*resgain(12,30)", 16384); 
       // specify filter string and sampling frequency
    fdesign.bode (0.1, 1000, 1001); // make a bode plot
    fdesign.response ("step", 10.0) // plot step response
    \end{verbatim}

    \begin{verbatim}
    Example 3: Read DARM filter from online system, add a 60Hz comb 
               and test it on random noise
    FilterDesign fdesign (16384); 
    fdesign.frontend ("H2LSC.txt", "DARM");
    fdesign.comb (60, 100); // add Q=100 comb
    TSeries test (Time (0), Interval (1./fdesign.getFSample()), 
                  10*16384, GaussNoise (fdesign.getFSample())); 
            // make some random noise
    fdesign.response (test) // plot filter response to random noise
    \end{verbatim}

    The filter types are: kLowPass for low pass, kHighPass for high
    pass, kBandPass for band pass and kBandStop for a band stop
    filter, respectively. When using a string specification, the 
    corresponding filter types are "LowPass", "HighPass", "BandPass"
    and "BandStop".

    FILTER FORMULA SPECIFICATION

    Filter formulas are written as a product of real numbers and
    designer functions. Optionally one can add a gain condition
    at the end. Designer functions ar of the form "f(a1,a2,...aN)"
    where "f" must be a supported function and "a1", .. "aN" must
    be the corresponding arguments. A complex number argument is 
    written as "real+i*imag" or "real+imag*i". Number arguments
    must be constants and expressions are not allowed. A vector
    of numbers is written as "[e1;e2;...eN]". The gain condition
    is introduce by the "/." symbol and it must be followed by the 
    "setgain(f,g)" where g is the desired gain and f is the 
    frequency at which it should be true. The multiplication 
    operator "*" is optional and can be omitted or replaced by a
    space.

    List of designer functions:
    \begin{verbatim}
    gain(g,gfmt='scalar')              multiply by g
    pole(f,g=1,pl='s')                 a simple pole
    pole(f,pl='s')                     a simple pole
    zero(f,g=1,pl='s')                 a simple zero
    zero(f,pl='s')                     a simple zero
    pole2(f,Q,g=1,pl='s')              a complex pole pair
    pole2(f,Q,pl='s')                  a complex pole pair
    zero2(f,Q,g=1,pl='s')              a complex zero pair
    zero2(f,Q,pl='s')                  a complex zero pair
    zpk([z1;...zN],[p1;...pM],g=1,pl)  a set of s-plane poles and zeros
    rpoly([z1;...zN],[p1;...pM],g=1)   a rational polynomial
    biquad(b0,b1,b2,a1,a2)             second order section
    sos(g,[b1,b2,a1,a2,...],fmt='s')   list of second order sections
    zroots([z1;...zN],[p1;...pM],g=1)  a set of z-plane poles and zeros
    direct([b0;...bN],[a1;...aM])      direct form
    ellip(type,order,rp,as,f1,f2=0)    elliptic filter
    butter(type,order,f1,f2=0)         butterworth filter
    cheby1(type,order,rp,f1,f2=0)      chebyshev type 1 filter
    cheby2(type,order,as,f1,f2=0)      chebyshev type 2 filter
    notch(f,Q,depth=0)                 notch filter
    resgain(f,q,height=0)              resonant gain filter
    comb(f,Q,depth=0,hN=0)             comb filter 
    remez(order,[...],[...],[...])     FIR filter according to McClellan-Parks
    firw(...)                          FIR filter with window method
    difference()                       difference filter
    limiter(type,l1,l2=0,l3=0)         limiter
    decimateBy2(stages,id=1)           decimate by factors of 2 (FIR)
    multirate(type,m1,m2,attn)         resample the time series
    mixer(fmix,phase)                  mixer
    linefilter(...)                    line removal
    frontend(file,module,sect=-1)      read front-end configuration
    \end{verbatim}
    where 
    \begin{verbatim}
    f             corner frequency,
    g             gain,
    gfmt          gain format
    Q             quality factor
    z             zero location
    p             pole location
    pl            pole/zero location
    a1,a2         denominator coefficents
    b0,b1,b2      numerator coefficents
    fmt           format of second order sections
    type          filter type
    order         filter order
    rp            passband ripple
    as            stopband attenuation
    f1            corner frequency
    f2            corner frequency
    depth         depth in dB
    height        depth in dB
    hN            Number of harmonics
    l1,l2,l3      limits (bounds and/or slew rate)
    stages        number of stages
    id            filter id
    m1,m2         resampling parameters (abs or rel)
    attn          attenuation in dB
    fmix          heterodyne frequency
    phase         phase
    file          file name
    module        filter module name
    sect          filter section
    \end{verbatim}
    For a more detailed description of the designer functions see
    the documentation of the corresponding method below.
    

    @memo Filter design
    @author Written July 2002 by Ed Daw, Masahiro Ito, Daniel Sigg
    @version 1.0
 ************************************************************************/
   class FilterDesign : public FilterParse {
   public:
      /** Constructs a filter design class. The filter design class
          keeps track of the sampling frequency and the current 
          filter.
          @memo Constructor.
          @param fsample Sampling rate
          @param name Name of filter
       ******************************************************************/
      explicit FilterDesign (double fsample = 1.0, 
                        const char* name = "filter");
      /** Constructs a filter design class. The current filter is 
          initialized with the specified string---see also 
          AddFilter(const char*).
          @memo Constructor.
          @param spec Filter design specification
          @param fsample Sampling rate
          @param name Name of filter
       ******************************************************************/
      explicit FilterDesign (const char* spec, double fsample = 1.0, 
                        const char* name = "filter");
      /** Copy constructor.
          @memo Copy constructor.
       ******************************************************************/
      FilterDesign (const FilterDesign& design);
      /** Destructs the filter design class.
          @memo Destructor.
       ******************************************************************/
      virtual ~FilterDesign();
      /** Assignment operator.
          @memo Assignment operator.
       ******************************************************************/
      FilterDesign& operator= (const FilterDesign& design);
      /** Initializs the filter design class with a new sampling rate.
          The output rate is set equal to the input rate and the
          current filter is reset to NULL. The prewarp flag is left
          unchanged.
          @memo Initialization.
          @param fsample Sampling rate
          @return void
       ******************************************************************/
      void init (double fsample = 1.0);
   
      /** Check if we have a trivial unity gain filter.
          @memo Check for unity gain filter.
          @return true if unity gain
       ******************************************************************/
      bool isUnityGain () const;
   
      /** Set the filter name.
          @memo Set name.
          @param name Name of filter
          @return void
       ******************************************************************/
      void setName (const char* name) {
         fName = name ? name : "filter"; }
      /** Get the filter name.
          @memo Get name.
          @return filter name
       ******************************************************************/
      const char* getName() const {
         return fName.c_str(); }
      /** Set the sampling frequency. The current filter gets reset.
          @memo Set sampling frequency.
          @param fsample Sampling rate
          @return void
       ******************************************************************/
      void setFSample (double fsample) {
         init (fsample); }
      /** Get the (intput) sampling frequency.
          @memo Get sampling frequency.
          @return Sampling rate
       ******************************************************************/
      double getFSample() const {
         return fFSample; }
      /** Get the output sampling frequency.
          @memo Get output sampling frequency.
          @return Sampling rate
       ******************************************************************/
      double getFOut() const {
         return fFOut; }
      /** Get the heterodyne flag. This flag is set to false initially
          and will be set to true if one of the filters is a mixer.
          This is then an indication that the output time series will
          be complex for a real input time series.
          @memo Get heterodyne flag.
          @return Sampling rate
       ******************************************************************/
      bool getHeterodyne() const {
         return fHeterodyne; }
      /** Set the prewarping flag
          @memo Set prewarping flag
          @param prewarping flag
          @return void
       ******************************************************************/
      void setPrewarp (bool set = true) {
         fPrewarp = set; }
      /** Is prewarping enabled?
          @memo Get prewarping flag
          @return true if prewarping is enabled
       ******************************************************************/
      bool getPrewarp() const {
         return fPrewarp; }
      /** Get the filter specification.
          @memo Get filter specification.
          @return Filter specification
       ******************************************************************/
      const char* getFilterSpec() const {
         return fFilterSpec.c_str(); }
      /** Get pad returned by the last call to a plotting function.
          @memo Get plot pad.
          @return plot pad
       ******************************************************************/
      ligogui::TLGMultiPad* getPad() const {
         return fPad; }
   
   
      /** Get the filter. Returns the identity filter, if no filter 
          has been added.
          @memo Get the filter
          @return current filter
       ******************************************************************/
      Pipe& get() {
         return *fFilter; }
      /** Get the filter. Returns the identity filter, if no filter 
          has been added.
          @memo Get the filter (const)
          @return current filter
       ******************************************************************/
      const Pipe& get() const {
         return *fFilter; }
      /** Get the filter. Returns the identity filter, if no filter 
          has been added.
          @memo Get the filter
          @return current filter
       ******************************************************************/
      Pipe& operator() () {
         return get(); }
      /** Get the filter. Returns the identity filter, if no filter 
          has been added.
          @memo Get the filter (const)
          @return current filter
       ******************************************************************/
      const Pipe& operator() () const {
         return get(); }
      /** Get a copy of the filter. User must delete it! Returns the 
          identity filter, if no filter has been added.
          @memo Get a copy of the filter
          @return copy of current filter
       ******************************************************************/
      Pipe* copy() const;
      /** Release the current filter and return it. Returns the identity 
          filter, if no filter has been added. After this call the 
          current filter is set back to an identity filter.
          @memo Get the current filter
          @return current filter
       ******************************************************************/
      Pipe* release();
      /** Resets the current filter to identity.
          @memo Reset the current filter
          @return void
       ******************************************************************/
      void reset();
      /** Set the filter.
          @memo Set the filter
          @param filter New current filter
          @param resampling Resampling factor
          @param heterodyne Filter is heterodyning
       ******************************************************************/
      void set (const Pipe& filter, double resampling = 1.0, 
               bool heterodyne = false);
      /** Add a filter to the current one.
          @memo Add a filter
          @param  filter Filter to be added
          @param heterodyne Filter is heterodyning
          @return pointer to new current filter (owned by design class)
       ******************************************************************/
      bool add (const Pipe& filter, double resampling = 1.0, 
               bool heterodyne = false);
   
      /** Make a filter from a descriptive string and add it to the
          current one.
          @memo Add a filter
          @param spec New filter specification
          @return true if successful
       ******************************************************************/
      bool filter (const char* formula);
   
      /** Multiply the filter by the specified gain. The gain can be 
          specified as a scalar or in dB. The default is scalar.
          @memo Add a gain
          @param g gain.
          @param format scalar or dB
          @return true if successful
       ******************************************************************/
      bool gain (double g, const char* format = "scalar");
   
      /** Make a simple pole. 
          Poles and zero can be specifed in the s-plane using units of 
          rad/s (set plane to "s") or units of Hz ("f"). In both cases 
          the frequency is expected to be POSITIVE. Alternativly, one can 
          also specify normalized poles and zeros ("n"). Normalized poles 
          and zeros are expected to have POSITIVE frequencies and their 
          respective low and high frequency gains are set to 1---unless 
          they are located a 0Hz. The default loaction is the s-plane.
          For the s-plane the formula is $pole(f0) = 1/(s + w0)$ ["s"] 
          or $pole(f0) = 1/(s + 2 pi f0)$ ["f"]. For a normalized pole 
          ["n"] the formula is $pole(f0) = 1/(1 + i f/f0)$; $f0 > 0$ and
          $pole(f0) = 1/(i f)$; $f0 = 0$. 
          @memo Add pole
          @see IIRFilter.hh
          @param f Pole frequency.
          @param g gain.
          @param plane location where poles/zeros are specified
          @return true if successful
       ******************************************************************/
      bool pole (double f, double gain, const char* plane = "s");
      bool pole (double f, const char* plane = "s");
      /** Make a simple zero. 
          Poles and zero can be specifed in the s-plane using units of 
          rad/s (set plane to "s") or units of Hz ("f"). In both cases 
          the frequency is expected to be POSITIVE. Alternativly, one can
          also specify normalized poles and zeros ("n"). Normalized poles 
          and zeros are expected to have POSITIVE frequencies and their 
          respective low and high frequency gains are set to 1---unless 
          they are located a 0Hz. The default loaction is the s-plane.
          For the s-plane the formula is zero(f0) = (s + w0)$ ["s"] 
          or $zero(f0) = (s + 2 pi f0)$ ["f"]. For a normalized zero 
          ["n"] the formula is $zero(f0) = (1 + i f/f0)$; $f0 > 0$ and
          $zero(f0) = (i f)$; $f0 = 0$.
          @memo Add zero
          @see IIRFilter.hh
          @param f Zero frequency.
          @param g gain.
          @param plane location where poles/zeros are specified
          @return true if successful
       ******************************************************************/
      bool zero (double f, double gain, const char* plane = "s");
      bool zero (double f, const char* plane = "s");
      /** Make a complex pole pair.
          Poles and zero can be specifed in the s-plane using units of 
          rad/s (set plane to "s") or units of Hz ("f"). In both cases 
          the frequency is expected to be POSITIVE. Alternativly, one can
          also specify normalized poles and zeros ("n"). Normalized poles 
          and zeros are expected to have POSITIVE frequencies and their 
          respective low and high frequency gains are set to 1---unless 
          they are located a 0Hz. The default loaction is the s-plane.
          For the s-plane the formula is 
          $pole2(f0,Q)=1/(s^2 + w0 s / Q + w0^2)$ ["s"] or 
          $pole2(f0,Q)=1/(s^2 + 2 pi f0 s / Q + (2 pi f0)^2)$ ["f"]. 
          For a normalized complex pole pair ["n"] the formula is 
          $pole2(f0) = 1/(1 + i f/f0 + (f/f0)^2)$ with $f0 > 0$.
          The quality factor Q must be greater than 0.5.
          @memo Add a complex pole pair
          @see IIRFilter.hh
          @param f Pole frequency.
          @param Q Quality factor.
          @param g gain.
          @param plane location where poles/zeros are specified
          @return true if successful
       ******************************************************************/
      bool pole2 (double f, double Q, double gain,
                 const char* plane = "s");
      bool pole2 (double f, double Q, const char* plane = "s");
      /** Make a complex zero pair.
          Poles and zero can be specifed in the s-plane using units of 
          rad/s (set plane to "s") or units of Hz ("f"). In both cases 
          the frequency is expected to be POSITIVE. Alternativly, one can
          also specify normalized poles and zeros ("n"). Normalized poles 
          and zeros are expected to have POSITIVE frequencies and their 
          respective low and high frequency gains are set to 1---unless 
          they are located a 0Hz. The default loaction is the s-plane.
          For the s-plane the formula is 
          $zero2(f0,Q)=(s^2 + w0 s / Q + w0^2)$ ["s"] or 
          $zero2(f0,Q)=(s^2 + 2 \pi f0 s / Q + (2 \pi f0)^2)$ ["f"]. 
          For a normalized complex pole pair ["n"] the  formula is 
          $pole2(f0) = (1 + i f/f0 + (f/f0)^2)$ with $f0 > 0$.
          The quality factor Q must be greater than 0.5.
          @memo Add a complex zero pair
          @see IIRFilter.hh
          @param f Pole frequency.
          @param Q Quality factor.
          @param g gain.
          @param plane location where poles/zeros are specified
          @return true if successful
       ******************************************************************/
      bool zero2 (double f, double Q, double gain,
                 const char* plane = "s");
      bool zero2 (double f, double Q, const char* plane = "s");
      /** Make an filter from zeros and poles. 
          Poles and zero can be specifed in the s-plane using units of 
          rad/s (set plane to "s") or units of Hz ("f"). In both cases 
          the real part of the roots are expected to be NEGATIVE. 
          Alternativly, one can also specify normalized poles and zeros
          ("n"). Normalized poles and zeros are expected to have POSITIVE
          real parts and their respective low and high frequency gains
          are set to 1---unless they are located a 0Hz. The default 
          loaction is the s-plane. 
   
          For the s-plane the formula is
          $$zpk(s) = k \frac{(s - s_1)(s - s_2) \cdots}{(s - p_1)(s - p_2) \cdots}$$
          with $s_1$, $s_2$,... the loactions of the zeros and 
          $p_1$, $p_2$,... the loaction of the poles. By replacing 
          s_1->$2 \pi fz_1$and p_1->$2 \pi fp_1$ one obatines the "f" 
          representation. For normalize poles and zeros ["n"] one uses 
          poles of the form $pole(f0) = 1/(1 + i f/f0)$; $f0 > 0$ and 
          $pole(f0) = 1/(i f)$; $f0 = 0$. The zeros are then of the form
          $zero(f0) = (1 + i f/f0)$; $f0 > 0$ and $zero(f0) = (i f)$; 
          $f0 = 0$. Poles and zeros with a non-zero imaginary part must 
          come in pairs of complex conjugates.
   
          @memo Add zero-pole-gain (zpk)
          @see IIRFilter.hh
          @param nzeros Number of zeros
          @param zero Array of zeros
          @param npoles Number of poles
          @param pole Array of poles
          @param gain Gain
          @param plane location where poles/zeros are specified
          @return true if successful
       ******************************************************************/
      bool zpk (int nzeros, const dComplex* zero, 
               int npoles, const dComplex* pole, double gain,
               const char* plane = "s");
      bool zpk (int nzeros, const fComplex* zero, 
               int npoles, const fComplex* pole, double gain,
               const char* plane = "s");
   
      /** Make a filter from a rational polynomial in s. 
          A rational polynomial in s is specified by the polynomial 
          coefficients in the numerator and the denominator in descending 
          order of s. 
   
          The formula is
          $$zpk(s) = k \frac{a_n s^{n_z} + a_{n-1} s^{n_z - 1} \cdots}
                            {b_n s^{n_p} + b_{n-1} s^{n_p - 1} \cdots}$$
          where $a_n$, $a_{n-1}$,..., $a_0$ are the coeffciients of the 
          polynomial in the numerator and $b_n$, $b_{n-1}$,..., $b_0$ are
          the coefficients of the polynomial in the denominator.
          The polynomial coefficients have to be real.
   
          @memo Add a rational polynomial (rpoly)
          @see IIRFilter.hh
          @param nnumer Number of coefficients in numerator
          @param numer Array of numerator coefficents
          @param ndenom Number of coeffcients in denominator
          @param denom Array of denominator coeffcients
          @param gain Gain
          @return true if successful
       ******************************************************************/
      bool rpoly (int nnumer, const double* numer, 
                 int ndenom, const double* denom, double gain);
      bool rpoly (int nnumer, const float* numer, 
                 int ndenom, const float* denom, double gain);
   
      /** Make an filter from the biquad coefficients and add it to the 
          current one.
          @memo Add a second order section
          @see IIRdesign.hh
          @param b0 b0
          @param b1 b1
          @param b2 b2
          @param a1 a1
          @param a2 a2
          @return true if successful
       ******************************************************************/
      bool biquad (double b0, double b1, double b2, double a1, double a2);
   
      /** Make a IIR filter from a list of second order sections. If the
          format is 's' (standard), the coefficents are ordered like:
          $$ gain, b1_1, b2_1, a1_1, a2_1, b1_2, b2_2, a1_2, a2_2,... $$
          whereas for the the format 'o' (online) the order is
          $$ gain, a1_1, a2_1, b1_1, b2_1, a1_2, a2_2, b1_2, b2_2,... $$
          The number of coefficients must be 4 times the number of second 
          order sections plus one.
          @memo Add a list of second order sections
          @see IIRdesign.hh
          @param nba Number of coefficients
          @param ba List of coefficients
          @param format Coefficient format
          @return true if successful
       ******************************************************************/
      bool sos (int nba, const double* ba, const char* format = "s");
   
      /** Make a IIR filter from the list of poles and zeros in the 
          z-plane. To be stable the z-plane poles must lie within the 
          unity cirlce.
          @memo Add a filter from z-plane roots
          @see IIRdesign.hh
          @param nzeros Number of zeros
          @param zero Array of zeros
          @param npoles Number of poles
          @param pole Array of poles
          @param gain Gain
          @return true if successful
       ******************************************************************/
      bool zroots (int nzeros, const dComplex* zero,
                  int npoles, const dComplex* pole, double gain = 1.0);
      bool zroots (int nzeros, const fComplex* zero,
                  int npoles, const fComplex* pole, double gain = 1.0);
   
      /** Make a filter from the direct form. 
          The direct form can be written as
          $$H(z)=\Sum_{k=0}^{nb}b_k z^{-k} / (1-\Sum_{k=1}^{na}a_k z^{-k})$$
    
          Cascaded second order sections are formed by finding the roots
          of the direct form. The specified coefficients are $b_0$, $b_1$,...,
          $b_{nb}$ for the numerator and $a_1$, $a_2$,..., $a_{na}$ for 
          the denominator. The argument $nb$ specifies the number of b 
          coeffcients supplied to the function minus 1, whereas $na$ is 
          exactly the number of coeffcients since $a_0$ is always 1 and 
          omitted from the list.
    
          Avoid the direct form since even fairly simple filters will 
          run into precision problems.
   
          @memo Add a filter from the direct form
          @param nb Number of coefficients in numerator
          @param b Array of b coefficents
          @param na Number of coeffcients in denominator
          @param a Array of denominator coeffcients exclusive a0
          @return true if successful
       ******************************************************************/
      bool direct (int nb, const double* b, int na, const double* a);
   
      /** Make an elliptic filter and add it to the current one.
          @memo Add an elliptic filter
          @see IIRdesign.hh
          @param type Filter type
          @param order Filter order
          @param rp Pass band ripple (dB)
          @param as Stop band attenuation (dB)
          @param f1 Pass band edge (Hz)
          @param f2 Another pass band edge (Hz)
          @return true if successful
       ******************************************************************/
      bool ellip ( Filter_Type type, int order, double rp, double as, 
                 double f1, double f2 = 0.0 );
      /** Make an chebyshev filter of type 1 and add it to the current 
          one.
          @memo Add a chebyshev filter
          @see IIRdesign.hh
          @param type Filter type
          @param order Filter order
          @param rp Pass band ripple (dB)
          @param f1 Pass band edge (Hz)
          @param f2 Another pass band edge (Hz)
          @return true if successful
       ******************************************************************/
      bool cheby1 ( Filter_Type type, int order, double rp, 
                  double f1, double f2 = 0.0 );
      /** Make an chebyshev filter of type 2 and add it to the current 
          one.
          @memo Add a chebyshev filter
          @see IIRdesign.hh
          @param type Filter type
          @param order Filter order
          @param as Stop band attenuation (dB)
          @param f1 Pass band edge (Hz)
          @param f2 Another pass band edge (Hz)
          @return true if successful
       ******************************************************************/
      bool cheby2 ( Filter_Type type, int order, double as, 
                  double f1, double f2 = 0.0 );
      /** Make an butterworth filter and add it to the current one.
          @memo Add a butterworth filter
          @see IIRdesign.hh
          @param type Filter type
          @param order Filter order
          @param f1 Pass band edge (Hz)
          @param f2 Another pass band edge (Hz)
          @return true if successful
       ******************************************************************/
      bool butter ( Filter_Type type, int order, double f1, 
                  double f2 = 0.0 );
      /** Make an notch filter and add it to the current one.
          @memo Add a notch filter
          @see IIRdesign.hh
          @param f0 Center frequency.
          @param Q Quality factor ( Q = (Center freq)/(Width) ).
          @param depth Depth of the notch (dB).
          @return true if successful
       ******************************************************************/
      bool notch ( double f0, double Q, double depth = 0.0 );
      /** Make a resonant gain filter and add it to the current one.
          @memo Add a resonant gain filter
          @see IIRdesign.hh
          @param f0 Center frequency.
          @param Q Quality factor ( Q = (Center freq)/(Width) ).
          @param height Height of the peak (dB).
          @return true if successful
       ******************************************************************/
      bool resgain ( double f0, double Q, double height = 30.0 );
      /** Make an comb filter and add it to the current one.
          @memo Add a comb filter
          @see IIRdesign.hh
          @param f0 Fundamental frequency.
          @param Q Quality factor ( Q = (Center freq)/(Width) ).
          @param amp Depth/height of notches/peaks (dB).
          @param N Number of harmonics.
          @return true if successful
       ******************************************************************/
      bool comb ( double f0, double Q, double amp = 0.0, int N = 0 );
   
      /** Design an FIR filter using the McClellan-Parks algorithm
          and add it to the current one.
          Design a optimal equiripple linear phase FIR filter from a 
          list of bands and the desired amplitudes in each band. 
          nBand defines the number of bands in which the response will
          be specified. 'Bands' contains a minimum and a maximum 
          frequency for each band. 'Func' contains the desired amplitude 
          for each band and 'Weight' contains an optional weighting 
          factor for each band. The resulting filter coefficients are 
          stored in the Filter object. The filter history is left 
          unmodified. 
          @memo  FIR filter design using the McClellan-Parks algorithm.
          @see FIRdesign.hh
          @param N      Number of filter coefficients.
          @param nBand  Number of bands.
          @param Bands  Band limit frequencies in Hz.
          @param Func   Desired response in each band.
          @param Weight Weighting factor for each band. If not specified, 
                        Weight is assumed to be 1.0 for each band.
          @return true if successful
       ******************************************************************/
      bool remez (int N, int nBand, const double* Bands, 
                 const double* Func, const double* Weight = 0);
   
      /** Design an FIR filter using the window method and add it to 
          the current one.
          Design a linear phase FIR filter using the window function 
          technique. The windowing functions that may be used are 
          "Rectangle", "Triangle", "Hamming", "Hanning", "Kaiser" or 
          "Cheby". Filter types that may be designed are "LowPass", 
          "HighPass", "BandPass" or "BandStop". N or dF will be 
          inferred from the other parameters if omitted or set 
          to zero in Kaiser and Chebyshev designs. The same is true of 
          Ripple if it is omitted or set to zero in Chebyshev designs. 
          The designed Filter coefficients replace the current 
          coefficients. The Filter history is not affected.
          @memo  Filter design from window function.
          @see FIRdesign.hh
          @param N      Number of filter coefficients.
          @param window Type of window on which to base the design.
          @param type   Type of filter.
          @param Flow   Transition frequency of a low-pass or 
                        high-pass filter, or lower edge of a band-pass 
                        or band-stop filter.
          @param Fhigh  Upper edge of a band-pass or band-stop filter.
          @param Ripple Desired attenuation for Kaiser filters or the 
                        desired pass-band ripple for Chebyshev filters 
                        (in dB).
          @param dF     Transition band width in Chebyshev filters.
          @return true if successful
       ******************************************************************/
      bool firw (int N, Filter_Type type, const char* window, 
                double Flow, double Fhigh = 0, double Ripple = 0, 
                double dF = 0);
   
      /** Make a difference filter and add it to the current one.
          Differential filter takes the difference between each 
          successive pair of elements.
          @see Difference.hh
          @memo Add a difference filter
          @return true if successful
       ******************************************************************/
      bool difference();
   
      /** Make a decimate-by-2 filter and add it to the current one.
          The DecimateBy2 class decimates a TSeries by 2^N. The 
          TSeries data are filtered before decimation to remove 
          aliasing. This class is essentially a wrapper of the C 
          code written for LIGO GDS by Peter Fritschel.
   
          DecimateBy2 is based on the DMT FilterBase class and as 
          such is can be treated as a generic DMT Filter. The 
          decimation may be applied to the TSeries with either the 
          apply method or the parentheses operator. These methods 
          return a complex TSeries if the input series is complex,
          and a single precision floating point TSeries for all other 
          input series data types.
   
          Four anti-aliasinglow-pass FIR filters are available, 
          each of which cuts of at 0.9 F_{Nyquist}. The options are 
          tabulated below:
          \begin{tabular}{r|r|l}
          ID & Order & Comments \\ \hline
          1 & 42 & Least-Squares design, ripple ~.015 (-0.2 dB) \\
          2 & 42 & Equiripple design, ripple ~0.06 dB \\
          3 & 22 & Least-squares design, ripple ~0.1 (-1 dB) \\
          4 & 82 & Least-squares design, ripple ~0.0006 (-0.01 dB) \\
       \end{tabular}
          The antialiasing filters add an effective delay of $N*O_F/2$,
          where N is the number of stages (the decimation factor 
          exponent) and $O_F$ is the filter order. This delay is NOT 
          added to the TSeries start time.
          @memo Add a decimation by 2^N.
          @see DecimateBy2.hh
          @param N number of stages
          @param FilterID filter type
          @return true if successful
       ******************************************************************/
      bool decimateBy2 (int N, int FilterID = 1);
   
      /** Make a line removal filter and add it to the current one.
          The LineFilter class containes methods to track and remove 
          quasi-monochromatic lines. A filter id of 0 corresponds
          to a pure "comb" filter, a filter if of 1 corresponds to 
          an optimal Wiener filter with 1-st method of noise estimation 
          (default), and a filter id of -1 
          @memo Add a line removal filter
          @see LineFilter.hh
          @param f Approximate line frequency
          @param T time interval to remove lines.
          @param id Select filter ID (0/1), 1 by default
          @param nT number of subdivisions of time interval w
          @return true if successful
       ******************************************************************/
      bool linefilter (double f, double T = 0.0, int fid = 1, 
                      int nT = 1);
   
      /** Make a limiter filter and add it to the current one.
          The type can be "" for no limits, "val" for high/low limits on
          the sampled values, "sym" for a symmetric limits on the values, 
          "slew" for a slew rate limit and "val/slew" or "sym/slew" for 
          both. If the limiter type is "val", l1 is the lower bound and 
          l2 is the upper bound, for "sym" l1 is the upper bound and -l1
          is the lower bound, for "slew" l1 is the slew rate limit in 
          counts/s, for "val/slew" l1, l2 and l3 are the lower bound, 
          the upper bound and the slew rate limit, respectively, and for 
          "sym/slew" l1 and l2 are the symmetric limit and the slew rate 
          limit, respectively.
          @memo Add a limiter
          @see Limiter.hh
          @param type Limiter type
          @param l1 First limit
          @param l2 Second limit
          @param l3 Third limit
          @return true if successful
       ******************************************************************/
      bool limiter (const char* type, double l1, double l2 = 0, 
                   double l3 = 0);
   
      /** Make a resampling filter and add it to the current one.
          The filter type is either "abs" or "rel" depending if the 
          required resampling rate is specified realtive or absolute.
          For absolute rate specifications the first parameter is the 
          desired sampling rate and the second parameter is the maximally 
          allowed error. For a relative rate specifiaction the first 
          parameter is the interpolation factor and the second parameter
          is the decimation factor (both are integer parameters). The 
          third parameter is in both cases the required stop band 
          attenuation.
   
          @memo Add a resampling filter
          @see MultiRate.hh
          @param type Multirate type
          @param m1 First parameter (desired fS or interpolation factor)
          @param m2 Second parameter (error or decimation factor)
          @param attn Stopband attenuation
          @return true if successful
       ******************************************************************/
      bool multirate (const char* type, double m1, double m2 = 1E-3, 
                     double atten = 80);
   
      /** Make a mixer and add it to the current filter. This will 
          transform a real time series into a complex one. The mixer
          will heterodyne (multiply) a time series by 
          exp(2*pi*I*fc*t+phase).
          @memo Add a mixer
          @see Mixer.hh
          @param fmix Mixer frequency
          @param phase Initial phase
          @return true if successful
       ******************************************************************/
      bool mixer (double fmix, double phase = 0);
   
      /** Read filter coefficients from front-end file, make filter 
          and add it to the current one.
          @memo Add a filter from a configuaryion file
          @param file Front-end configuration file
          @param module Module name or number
          @param section Filter section or -1 for all
          @return true if successful
       ******************************************************************/
      bool frontend (const char* file, const char* module, 
                    int section = -1);
   
      /** Set the filter gain at specified frequency.
          @memo Set filter gain
          @param f Frequency 
          @param gain Set point
          @return true if successful
       ******************************************************************/
      bool setgain (double f, double gain);
   
      /** Form the closed loop response of the current filter.
          @memo Closed loop response
          @param k Additional gain: 1/(1+k*G(f)) 
          @return true if successful
       ******************************************************************/
      bool closeloop (double k = 1.0);
   
      /** The transfer coefficient of the filter at the specified 
          frequency is calculated and returned as a complex number. 
          @memo Get a transfer coefficent of a Filter.
          @param coeff a complex number representing the Filter response 
                       at the specified frequency (return)
          @param F Frequency at which to sample the transfer function.
          @return true if successful       
       ******************************************************************/
      bool Xfer (fComplex& coeff, double f) const;
      /** The transfer function of the filter at the specified frequency 
          points is calculated and returned as a complex array. 
          The frequency points are user supplied. The return array
          must be at least of length points.
          @memo Get the transfer function of a Filter.
          @param tf Transfer function (return)
          @param freqs Frequency points
          @param points Number of points.
          @return true if successful       
       ******************************************************************/
      bool Xfer (fComplex* tf, const float* freqs, int points) const;
      /** The transfer function of the filter in the specified frequency 
          interval is calculated and returned as a complex frequency 
          series. 
          @memo Get the transfer function of a Filter.
          @param fseries a complex FSeries containing the Filter response 
                         at each frequency step (return)
          @param Fmin Minimum frequency at which to sample the 
                      transfer function.
          @param Fmax Maximum frequency at which to sample the 
                      transfer function.
          @param dF Frequency step.
          @return true if successful       
       ******************************************************************/
      bool Xfer (FSeries& fseries, double Fmin = 0.0, double Fmax = 0.0, 
                double dF = 1.0) const;
      /** The transfer function of the filter at the specified frequency 
          points is calculated and returned as a complex array. 
          The 'sweep type' can be "log" for logarithmically spaced
          frequency points or "linear" for linear spacing. The specifed 
          arrays must be at least of length points.
          @memo Get the transfer function of a Filter.
          @param freqs Frequency points (return)
          @param tf Transfer function (return)
          @param fstart Minimum frequency at which to sample the 
                        transfer function.
          @param fstop Maximum frequency at which to sample the 
                       transfer function.
          @param points Number of points.
          @param type Type of frequency spacing.
          @return true if successful       
       ******************************************************************/
      bool Xfer (float* freqs, fComplex* tf, 
                double fstart, double fstop = 0.0, int points = 101,
                const char* type = "log") const;
      /** The transfer function of the filter in the specified frequency 
          interval is calculated using a swept sine. This method works
          with all filters but may take significantly longer than the
          direct method implemented for FIR and IIR filters.
          @memo Get the transfer function of a Filter.
          @param freqs Frequency points (return)
          @param tf Transfer function (return)
          @param param Swept sine parameters
          @return true if successful       
       ******************************************************************/
      bool Xfer (float* freqs, fComplex* tf,
                const SweptSine& param) const;
   
      /** Compute the response of the filter. The reqested reponse type
          must be either "step" for a step response, "impulse" for
          an impulse response, or "ramp" for a ramp response.
          @memo Compute filter response.
          @param output Time series (return)
          @param type Response type 
          @param duration Length of returned time series.
          @return time series of response       
       ******************************************************************/
      bool response (TSeries& output, const char* type = "step",
                    const Interval& duration = 1.0) const;
      /** Compute the response of the filter.
          @memo Compute filter response.
          @param output Time series (return)
          @param func Signal function 
          @param duration Length of returned time series.
          @return time series of response       
       ******************************************************************/
      bool response (TSeries& output, const Chirp& func,
                    const Interval& duration = 1.0) const;
      /** Compute the response of the filter. 
          @memo Compute filter response.
          @param output Time series (return)
          @param input Input time series 
          @return time series of response       
       ******************************************************************/
      bool response (TSeries& output, const TSeries& input) const;
   
      /** Bode plot of filter. Make a bode plot of the filter.
          @memo Bode plot.
          @param fstart Minimum frequency at which to sample the 
                        transfer function.
          @param fstop Maximum frequency at which to sample the 
                       transfer function.
          @param points Number of points.
          @param type Type of frequency spacing.
          @return true if successful       
       ******************************************************************/
      bool bode (double fstart, double fstop = 0.0, int points = 101,
                const char* type = "log") const;
      /** Bode plot of filter. Make a bode plot of the filter.
          @memo Bode plot.
          @param freqs.
          @param points Number of points.
          @param type Type of frequency spacing.
          @return true if successful       
       ******************************************************************/
      bool bode (const float* freqs, int points = 101) const;
      /** Bode plot of filter. Make a bode plot of the filter using a
          swept sine method. This method works with all filters but may 
          take significantly longer than the direct method implemented 
          for FIR and IIR filters.
          @memo Bode plot.
          @param param Swept sine parameters
          @param freqs Frequency points (return)
          @param tf Transfer function (return)
          @return true if successful       
       ******************************************************************/
      bool bode (const SweptSine& param) const;
   
      /** Plot the response of the filter. The reqested reponse type
          must be either "step" for a step response, "impulse" for
          an impulse response, or "ramp" for a ramp response.
          @memo Plot filter response.
          @param type Response type 
          @param duration Length of returned time series.
          @return true if successful       
       ******************************************************************/
      bool resp (const char* type = "step",
                const Interval& duration = 1.0) const;
      /** Plot the response of the filter.
          @memo Plot filter response.
          @param func Signal function 
          @param duration Length of returned time series.
          @return true if successful       
       ******************************************************************/
      bool resp (const Chirp& func,
                const Interval& duration = 1.0) const;
      /** Plot the response of the filter. 
          @memo Plot filter response.
          @param input Input time series 
          @return true if successful       
       ******************************************************************/
      bool resp (const TSeries& input) const;
   
      /** Interactive design of a filter. Not yet implemented.
          @memo Filter wizard.
          @return true if a new filter was designed       
       ******************************************************************/
      bool wizard();
   
   private:
      /// Input sampling frequency
      double		fFSample;
      /// Output sampling frequency
      double		fFOut;
      /// Heterodyne flag
      bool		fHeterodyne;
      /// Current filter
      Pipe*		fFilter;
      /// Use prewarping
      bool		fPrewarp;
      /// Name
      std::string	fName;
      /// Pad returned by plot functions
      mutable ligogui::TLGMultiPad*	fPad;
      /// filter specifiction
      std::string	fFilterSpec;
   
   protected:
      // plot bode
      bool plotbode (const float* freqs, const fComplex* tf,
                    int points) const;
      // plot time series
      bool plotts (const TSeries& ts) const;
   };


#endif // _LIGO_FILTERDESIGN_H
